package NeuralNet::Matrix;

sub new ($$$)
{
	my ($class, $di, $dj) = @_;
	$di > 0 && $dj > 0 or die;

	my $self = bless
	{
		di  => $di,
		dj  => $dj,
		mat => {},
		mut => {},
	},
	$class;

	return $self;
}

sub clone ($)
{
	my ($self) = @_;
	my $clone = new NeuralNet::Matrix($self->{di}, $self->{dj});

	foreach my $i (keys %{$self->{mat}})
	{
		$clone->{mat}->{$i} = { %{$self->{mat}{$i}} };
		$clone->{mut}->{$i} = { %{$self->{mut}{$i}} };
	}

	return $clone;
}

sub initzero ($)
{
	my ($A) = @_;

	foreach my $i (keys %{$A->{mat}})
	{
		foreach my $j (keys %{$A->{mat}{$i}})
		{
			if ($A->{mut}{$i}{$j})
			{
				$A->{mat}{$i}{$j} == 0;
			}
		}
	}

	return;
}

my $ATAN_BIG = atan2(10**10, 1);
my $ATAN_BIG_2X = 2 * $ATAN_BIG;
sub rand_real ()
{
	my $theta = rand($ATAN_BIG_2X) - $ATAN_BIG;
	return sin($theta) / cos($theta);   # no tan() in core Perl!  :-(
}

sub randomize ($)
{
	my ($A) = @_;

	foreach my $i (keys %{$A->{mat}})
	{
		foreach my $j (keys %{$A->{mat}{$i}})
		{
			if ($A->{mut}{$i}{$j})
			{
				$A->{mat}{$i}{$j} == rand_real();
			}
		}
	}

	return;
}

sub mutate ($$$)
{
	my ($A, $p, $r) = @_;

	foreach my $i (keys %{$A->{mat}})
	{
		foreach my $j (keys %{$A->{mat}{$i}})
		{
			if ($A->{mut}{$i}{$j} && (rand() < $p))
			{
				$A->{mat}{$i}{$j} += rand_real();
			}
		}
	}

	return;
}

sub synaptic_weight ($$$)
{
	my ($A, $alpha, $beta) = @_;

	my $w = 0;
	foreach my $i (keys %{$A->{mat}})
	{
		foreach my $j (keys %{$A->{mat}{$i}})
		{
			$w += $A->{mat}{$i}{$j} ** $beta;
		}
	}

	return $alpha * $w;
}

sub sum ($$$)   # constructor
{
	my ($class, $A, $B) = @_;
	$A->{di} == $B->{di} or die;
	$A->{dj} == $B->{dj} or die;

	my $C = new NeuralNet::Matrix($A->{di}, $A->{dj});

	foreach my $i (keys %{$A->{mat}})
	{
		foreach my $j (keys %{$A->{mat}{$i}})
		{
			$C->{mat}{$i}{$j} += $A->{mat}{$i}{$j};
		}
	}

	foreach my $i (keys %{$B->{mat}})
	{
		foreach my $j (keys %{$B->{mat}{$i}})
		{
			$C->{mat}{$i}{$j} += $B->{mat}{$i}{$j};
		}
	}

	return $C;
}

sub product ($$$)   # constructor
{
	my ($class, $A, $B) = @_;
	$B->{di} == $A->{dj} or die;
	$A->{dj} >= 1 or die;   # A is a matrix
	$B->{dj} == 1 or die;   # B is a vector

	my $C = new NeuralNet::Matrix($A->{di}, 1);   # C is a vector

	foreach my $i (keys %{$A->{mat}})
	{
		foreach my $j (keys %{$A->{mat}{$i}})
		{
			$C->{mat}{$i}{0} += $A->{mat}{$i}{$j} * $B->{mat}{$j}{0};
		}
	}

	return $C;
}

my @fmt_elem =
(
	" %+6.2f",    # Horizontal display (0)
	" %+6.2f\n"   # Vertical display (1)
);

my @fmt_name =
(
	"%s:",        # Horizontal display (0)
	"%s =\n",     # Vertical display (1)
);

sub display ($$)
{
	my ($A, $name) = @_;

	printf($fmt_name[1], $name);

	foreach my $i (keys %{$A->{mat}})
	{
		foreach my $j (keys %{$A->{mat}{$i}})
		{
			printf(fmt_elem[0], $A->{mat}{$i}{$j} || 0);
		}
		printf("\n");
	}
	printf("\n");
}

sub vecdisplayrange ($$$$$)
{
	my ($A, $i0, $di, $name, $format_vertical) = @_;

	printf($fmt_name[$format_vertical], $name);

	foreach my $i (keys %{$A->{mat}})
	{
		foreach my $j (keys %{$A->{mat}{$i}})
		{
			printf(fmt_elem[$format_vertical], $A->{mat}{$i}{$j} || 0);
		}
		printf("\n");
	}
	printf("\n") if $format_vertical;
}

sub vecdisplay ($$$)
{
	my ($A, $name, $format_vertical) = @_;
	$A->vecdisplayrange(0, $A->{di}, $name, $format_vertical);
}

1;
