\Module{Neural Network}

[This header file is a conglomeration of several earlier header files.]



\Section{General Types}

[description]

   The ANSI C header file 'limits.h' is used here to decide which integer types
to use for various derived fixed-size aliases.

|#include <limits.h>


Shorter, clearer aliases necessary to program in C.

|typedef  unsigned char   uchar;
|typedef  unsigned short  ushort;
|typedef  unsigned int    uint;
|typedef  unsigned long   ulong;
|
|typedef  unsigned int    bool;


8-bit integer types.

|#if    (UCHAR_MAX == 0xFF)
|
|  typedef  uchar   uint8;
|  typedef   char    int8;
|
|#else
|
|  #error No definition for "uint8"
|
|#endif


16-bit integer types.

|#if    (UINT_MAX == 0xFFFF)
|
|  typedef  uint    uint16;
|  typedef   int     int16;
|
|#elif  (USHRT_MAX == 0xFFFF)
|
|  typedef  ushort  uint16;
|  typedef   short   int16;
|
|#else
|
|  #error No definition for "uint16"
|
|#endif


32-bit integer types.

|#if    (UINT_MAX == 0xFFFFFFFF)
|
|  typedef  uint    uint32;
|  typedef   int     int32;
|
|#elif  (ULONG_MAX == 0xFFFFFFFF)
|
|  typedef  ulong   uint32;
|  typedef   long    int32;
|
|#else
|
|  #error No definition for "uint32"
|
|#endif


Real number type.

 typedef float real;            WHY WON'T THIS WORK RIGHT???  CASTING PROBLEMS?
|typedef double real;


Useful constants.

|#define  TRUE   1
|#define  FALSE  0


Useful macros.  [**describe**]

|#define  NUM_ELEMENTS(x)  (sizeof(x) / sizeof((x)[0]))



\Section{Dynamic Arrays}

   The 'NEW' macro allocates a specified structure from the heap and may be
released with the 'Release' function.  Instead of writing this,

\S foo *p = (foo *) Alloc(sizeof(foo));

it is possible to write this:

\S foo *p = NEW(foo);

\ENTRY:  't' specifies a basic or derived type.

\EXIT:   'NEW(t)' is a pointer to a new object of type 't'.

|#define  NEW(t)  ((t *) Alloc(sizeof(t)))



\Section{Matrices and Vectors}

[OLD PARADIGM]

   A matrix is simply a dynamically allocated two-dimensional array of real
numbers; a vector is simply a dynamically allocated one-dimensional array of
real numbers.  Please see the array module for more details.

-typedef  real  **Matrix;
-typedef  real   *Vector;

[NEW PARADIGM]

   The new matrix paradigm is a one-dimensional array of matrix elements.
This array is capable of representing sparce matrices efficiently (credit to
R. McGehee for this data structure).  Each element contains a matrix value
$r$ and a matrix location $(i,j)$.  There is no particular order of the
elements.

|typedef struct
|{
|  int i;                    `Vertical coordinate.
|  int j;                    `Horizontal coordinate.
|  real r;                   `Value.
|  bool mut;                 `Mutable if true.
|}
|MatrixElement;
|
|typedef struct
|{
|  int di;                   `Height of 2D matrix.
|  int dj;                   `Width of 2D matrix (set to 1 for vector).
|  int n;                    `Current number of participating elements.
-  MatrixElement m[];        `di*dj slots for participating elements.
|  MatrixElement m[0];       `di*dj slots for participating elements.
|}
|Matrix;

   A vector is similar, but does not have a horizontal component.

|typedef struct
|{
|  int i;                    `Vertical coordinate.
|  real r;                   `Value.
|  bool mut;                 `Mutable if true.
|}
|VectorElement;
|
|typedef struct
|{
|  int di;                   `Height of 1D matrix.
|  int n;                    `Current number of participating elements.
-  VectorElement m[];        `di slots for participating elements.
|  VectorElement m[0];       `di slots for participating elements.
|}
|Vector;



\Section{Configuration}

   Variable matrix codes.  These are used to distinguish among constant zero,
constant one, and variable (e.g., mutatable) elements of the weight matrix and
threshold vectors.

|#define  VAR_MUTATE  (real)(-1.0)
|#define  VAR_ZERO    (real)( 0.0)
|#define  VAR_ONE     (real)( 1.0)

   Matrix template codes.  These are similar to the variable matrix codes in
that they codify types of states, but are different in that they are arbitrary
integers.  For speed in switch statements, these numbers should be consecutive
starting with zero (so the compiler can do a table-lookup and jump).  However,
nothing will break if this is not the case.

@@ Will VAR_xxxx become obsolete?

|#define  TEMPLATE_0  0
|#define  TEMPLATE_1  1
|#define  TEMPLATE_X  2



   Main configuration structure.  Here is held all of the data extracted from
the configuration file and from the command line.  Many of the substructures
are allocated dynamically at run-time based on neuron counts, population size,
and so forth.

   The model weight matrix 'M' and threshold vector 'B' are obtained from the
configuration file and are used in preparing new organisms.

|typedef struct
|{
|  ulong     seed;                 `Random number generator seed.
|
|  int       nn;                   `Total number of neurons.
|  int       ni;                   `Number of Input neurons.
|  int       nh;                   `Number of Hidden-layer neurons.
|  int       no;                   `Number of Output neurons.
|
|  int       ploidy;               `Ploidy of organisms (1, 2, etc.)
|  int       ndemes;               `Number of populations (demes).
|  int       npop;                 `Number of organisms (population size).
|  long      ngen;                 `Number of generations.
|  int       nmut;                 `Number of mutations per generation.
|  int       niter;                `Number of iterations for organism.
|
|  Matrix    *M;                   `Weight matrix.
|  Vector    *B;                   `Threshold vector.
|
|  int       ntrain;               `Number and array of training sets.
|  Vector    **train;
|
|  int       ntest;                `Number and array of test sets.
|  Vector    **test;
|
|  bool      initzero;             `Initialization mode (T = zero, F = rand).
|  bool      diffeq;               `Differential equations matrix flag.
|  bool      normall;              `Normalize all neurons flag.
|
|  bool      errorweight;          `Consider neuron weight in error calculation.
|  real      walpha;               `Weight alpha --- scalar multiplier.
|  real      wbeta;                `Weight beta --- exponent.
|}
|Config;


   The configuration structure is kept globally accessible so that it doesn't
need to be passed around to dozens of routines (previous versions of the
program passed a pointer in, and it got pretty ugly).

|extern Config c;



\Section{Organism Operations}

   Ploid structure.  A haploid organism consists of a single set of weight,
state, and threshold matrices.  A diploid organism consists of a pair of sets.

[** Perhaps X belongs OUTSIDE this structure...........examine ramifications.]

|typedef struct
|{
|  Matrix *M;                `Weight matrix.
|  Vector *X;                `State vector.
|  Vector *B;                `Threshold (boundary) vector.
|}
|Ploid;


   Organism structure.  Haploid and diploid organisms are constructed by
setting the ploidy to 1 or 2, respectively, and allocating a corresponding
number of ploid structures (defined above).

|typedef struct _Org
|{
|  struct _Org  *link;       `Pointer to next organism in the chain.  Used
|                            `only in the available list.
|
|  int     ploidy;           `Ploidy of organism (1, 2, etc.)
|  Ploid   *ploid;           `Poly-ploidic genetic makeup array.
|
|  real    edist;            `Distance (error) from ideal output set.
|  real    wdist;            `Weight distance from empty network.
|  real    dist;             `Total distance (sum of distance components).
|}
|Org;


   Aliases are created here for the haploid and diploid case.  The ploidy of an
organism is set to either HAPLOID (1) or DIPLOID (2).

|#define  HAPLOID   1        `Organism ploidy identifiers.
|#define  DIPLOID   2


   For the case of a diploid organism, aliases are created here for the
maternal and paternal portions.  The 'ploid' array in the organism structure
(defined above) is indexed by MATERNAL (0) or PATERNAL (1).

|#define  MATERNAL  0        `Array indexes for the two halves of a diploid
|#define  PATERNAL  1        `organism.



\Section{Error Handling}


Assert macro.  Can be enabled or disabled.  The cost is relatively low when
enabled, and zero when disabled.

|#if ASSERT
|
|  #define  Assert(expr)  if (!(expr)) AssertFunc(__FILE__, __LINE__, #expr)
|
|#else
|
|  #define  Assert(expr)
|
|#endif


Checkpoint macro.  Used only a couple of times in debugging.  Will not appear
in any final version of the program.

|#define  CHECKPOINT()  printf("Checkpoint:  %s(%d)\n", __FILE__, __LINE__);



\Section{Random Number Generation}

[**All of these would be nicer as functions. A bit slower, but a lot cleaner.]

   The 'RAND_INT_RANGE' macro produces random integers in a specified closed
interval $[min,max]$.

|#define  RAND_INT_RANGE(min, max)  (int) \
|  (min + (RandInteger() >> 16) * (long)(max - min + 1) / 65536L)


   The 'RAND_RANGE' macro produces random real numbers in a specified interval.
Since the function 'Rand' returns a number in the interval $[0,1)$, this macro
returns a number in the interval $[min,max)$.

|#define  RAND_RANGE(min, max)  (min + (Rand() * (max - min)))


   The 'RAND_REAL' macro produces random numbers over the range of real
numbers, with a bias toward small numbers.  A random number is chosen in the
range $-\pi/2$ to $\pi/2$, then the tangent of that number is calculated.  In
order to avoid infinitely large numbers (that is, practically speaking, machine
overflow), we use the arctangent of a very large number rather than exactly
$\pi/2$.

|#define  ATAN_BIG      1.570796326695       ` $atan(10^{10})$
|#define  RAND_REAL()   (real) tan(RAND_RANGE(-ATAN_BIG, ATAN_BIG))



\Section{Formatting constants}

Arrays, vectors, and perhaps some other objects can be displayed either
horizontally or vertically.

|#define  HDISPLAY  0
|#define  VDISPLAY  1


\Module{Neural Network}

This is a program for research on artificial neural nets---specifically for
biological applications---though the program is general and has other uses.  The
neural net has a \E{dynamic structure} comprising the instantaneous state of
each neuron.  This dynamic structure is simply a set of real numbers, one number
representing the \E{activity level} of each neuron in the net.  The dynamic
structure changes with time.  The neural net also has a \E{static structure}
comprising three things for each neuron:
(1)~a threshold, also a real number,
(2)~a set of real coefficients representing weighted stimuli arriving from other
    neurons in the net, and
(3)~a set of real coefficients representing weighted responses sent to other
    neurons in the net.
The static structure does not change with time---or at it least changes on a
much slower time scale.  Changing the static structure is called \E{training}.

Some of the neurons are designated \E{input neurons}; their activity levels are
simply the values of the external stimuli attached to them.  Others are
designated \E{output neurons}; their activity levels are taken as responses of
the entire neural net and connected to an external object or to another neural
net.  The remainder are called hidden neurons.

Among the $n$ neurons in a network are at most $n^2$ synapses and $n$
thresholds.  Synapses are represented by a matrix $M$ of weight values,
thresholds by a vector $b$, and the state of the system at time $t$ by a vector
$x_t$:
$$
M   = \pmatrix{ w_{11} & w_{12} & \ldots & w_{1n} \cr
                w_{21} & w_{22} & \ldots & w_{2n} \cr
                \vdots & \vdots & \ddots & \vdots \cr
                w_{n1} & w_{n2} & \ldots & w_{nn} \cr }
\,\,\,\,\,\,
b   = \pmatrix{ b_1    \cr
                b_2    \cr
                \vdots \cr
                b_n    \cr }
\,\,\,\,\,\,
x_t = \pmatrix{ x_{t,1} \cr
                x_{t,2} \cr
                \vdots  \cr
                x_{t,n} \cr }
\,.
$$

In these terms, the neural net becomes nothing more than the discrete dynamical
system defined by
$$ x_{t+1} = F(Mx_t + b) \,, $$
where
$$ F(x) = \pmatrix{ f(x_1) \cr
                    f(x_2) \cr
                    \vdots  \cr
                    f(x_n) \cr } $$
and $f(u)=\tanh(u/T)$ for hidden neurons and $f(u)=u$ for input and output
neurons.  $T$ is a simulated `temperature'---the degree of randomness in the
net.  The lower $T$, the sharper the non-linearity in $F$.  [Figure 1]

\midinsert
\input figtanh
{\narrower\narrower\noindent
%
% This needs to be a point or two smaller than the body of the document.
%
{\bf Figure 1.} The hyperbolic tangent function $f(u)=\tanh(u/T)$ for three
different values of $T$.  With $T=0$, the neural net becomes a digital device.

}
\endinsert

From this definition, $w_{ij}$ is the strength of the synapse by which neuron
$j$ stimulates neuron $i$.  The neural net is an iterated nonlinear map
on an $n$-dimensional real space $F:R^n \rightarrow R^n$.  All synchronous
neural nets that work from a sum of products and a threshold are special cases
of this model.

For example, consider a standard feed-forward neural net with four input
neurons, two hidden layers of three neurons each, and two output neurons.  This
is twelve neurons with only a small fraction of the $12\times12$ synapses
connected.  With neurons 1--4 as the inputs and neurons 11--12 as output, the
matrix has the structure.

\S                       1  2  3  4  5  6  7  8  9 10 11 12
\S                    1  1  -  -  -  -  -  -  -  -  -  -  -
\S                    2  -  1  -  -  -  -  -  -  -  -  -  -
\S                    3  -  -  1  -  -  -  -  -  -  -  -  -
\S                    4  -  -  -  1  -  -  -  -  -  -  -  -
\S                    5  X  X  X  X  -  -  -  -  -  -  -  -
\S                    6  X  X  X  X  -  -  -  -  -  -  -  -
\S                    7  X  X  X  X  -  -  -  -  -  -  -  -
\S                    8  -  -  -  -  X  X  X  -  -  -  -  -
\S                    9  -  -  -  -  X  X  X  -  -  -  -  -
\S                   10  -  -  -  -  X  X  X  -  -  -  -  -
\S                   11  -  -  -  -  -  -  -  X  X  X  -  -
\S                   12  -  -  -  -  -  -  -  X  X  X  -  -

Here 'X's represent real synapses (some value $w_{ij}\ne0$ takes the place of
each 'X'), dashes ('-') represent nonexistant synapses (where $w_{ij}=0$), and
'1's are used to hold values on input neurons.

As a second example, consider a Hopfield net. [Now describe this kind of
symmetric, fully connected network.]  [Also describe a competitive network.]

\H{Training}

[Include a section on genetic search explaining that the neural nets are trained
by simulated natural selection.  The program maintains a population of diploid
matrices.  Each cell of the operational matrix is the arithmetic mean (or
perhaps the sum) of the two haploid parts of the diploid matrix.  New matrices
are generated by recombination between  pairs of diploid matrices and by
mutation---using assortment among rows of the matrix and crossover within the
rows.  The genetic algorithm is independent of the rest of the program.]

\H{Format of the program}

The program is written in ANSI C in a document format.  Processed by a
\C{STAGE2} filter (Waite 1969) and further processed by the typesetting system
\TeX\ (Knuth 1986), it forms a document that describes the program in all
detail.  Processed by the utility 'd2d' it forms a program in normal C.  In the
makefile that accompanies this program, the command 'make program' creates an
executable version of the program and 'make document' creates a printable
document.  To prepare the document, one needs 'tex' and 'stage2'; to make the
program, one needs an \C{ansi} C compiler.  The document version is included in
the files labelled '*.dvi', and a version in a 300-dpi format is included in
files labelled '*.hp'.  To print a formatted copy of the document on a PC,
simply type the command 'copy /b *.hp lpt1'.

[We're not actually using \C{STAGE2}---we're using 'd2c'.]

[This last section does not belong here or in the makefile, but in an abstract
or something before the top-level table of contents.]

|#include <stdlib.h>
|#include <stdio.h>
-#include <conio.h>
|#include <string.h>
|#include <ctype.h>
|
|#include "neural.h"
|#include "protos.h"



\Section{Main}

This is the interface between the program and the operating system---in other
words the main entry point.  Command-line switches are defined in the
configuration module.  A file name is expected as the first non-switch
argument.

\ENTRY: 'argc' specifies the number of arguments to the program.\n
        'argv' is a list of pointers to the arguments.

\EXIT:  'main' returns a result code.\n
           0  No problems were encountered.\n
           1  Problems were encountered with the command line.\n
           2  A memory allocation failure occurred.

|#define  MAX_ORGANISMS  100
|
|int main (int argc, char *argv[])
|{
|  Org *u, *v, **(*pop);
|  char *configfile;
|  long gen;
|
|  int i, j, t;

|  printf("________________________________________"
|         "_______________________________________\n\n");

\H{Scan the command line for the configuration filename.}

|  configfile = (char *) 0;
|
|  for (i = 1; i < argc; i++)
|    if (argv[i][0] != '-')
|      configfile = argv[1];
|
|  if (!configfile)
|  {
|    printf("usage: neural [options] inputfile\n");
|    return 1;
|  }

\H{Read the configuration file.}

|  if (!ReadConfiguration(configfile) ||
|      !ReadArgConfiguration(argc, argv))
|    return 1;
|
|  DisplayConfiguration();
|
|  RandStart(c.seed);

\H{Create the initial population.}

|  pop = (Org ***) CreateArray2(c.ndemes, c.npop, sizeof(Org *));
|  for (i = 0; i < c.ndemes; i++)
|  for (j = 0; j < c.npop;   j++)
|  {
|    Org *o = OrgCreate(c.ploidy);
|    OrgInit(o);
|
|    if (c.initzero)
|      OrgZero(o);
|    else
|      OrgRand(o);
|
|    OrgEval(o);
|
|    pop[i][j] = o;                        `** I think I want OrgInsert here.
|  }

\H{Evolve the population until the user signals termination of the program.}

|  #define  KEY_NULL   0
|  #define  KEY_ESC    27
|  #define  KEY_SPACE  32
|
|  for (gen = 0; (!c.ngen || (gen < c.ngen)); gen++)
|  {
-    if (kbhit())                          `Check the keyboard and read a
|    if (0)                                `Check the keyboard and read a
|    {                                     `character if one is availble
|      int key = getch();                  `(MS-DOS) specific).
|
|      if (key == KEY_NULL)                `If the special null code is read,
|      {                                   `throw away the next character
|        getch();                          `without translation.  (MS-DOS)
|      }
|      else if (key == KEY_ESC)            `If the Escape key is pressed, halt.
|      {
|        break;
|      }
|      else if (key == KEY_SPACE)          `If the space bar is pressed,
|      {                                   `display the current weight and
|        MatDisplay(pop[0][0]->ploid[0].M, "M");` threshold matrices.
|        VecDisplay(pop[0][0]->ploid[0].B, "B", VDISPLAY);
\S                                   ^ Temporary hack
|      }
|    }
|
|    for (i = 0; i < c.ndemes; i++)
|    {
|      u = pop[i][RAND_INT_RANGE(0, c.npop - 1)];
|      v = OrgClone(u);
|      OrgMutate(v);
|      OrgEval(v);
|
|      if (OrgInsert(pop[i], c.npop, v) == 0)
|      {
|        if (c.errorweight)
|          printf("Gen:%9ld      "
|                 "e-dist:%10.6f     w-dist:%10.6f     dist:%10.6f\n",
|                 gen, (double)v->edist, (double)v->wdist, (double)v->dist);
|        else
|          printf("Generation: %9ld            "
|                 "e-dist: %13.6f\n",
|                 gen, (double)v->edist);
|
-        OrgHolisticTest(v);
|      }
|
|      if (gen % 5000 == 0)
|      {
|        if (c.errorweight)
|          printf("Gen:%9ld\n", gen);
|        else                                   `<<-- yuck, clean this up!
|          printf("Generation: %9ld\n", gen);
|
|        MatDisplay(v->ploid[0].M, "M");` threshold matrices.
|        VecDisplay(v->ploid[0].B, "B", VDISPLAY);
|      }
|    }
|  }

Report final matrix.

|  MatDisplay(pop[0][0]->ploid[0].M, "M");
|  VecDisplay(pop[0][0]->ploid[0].B, "B", VDISPLAY);
\S                             ^ Temporary hack

\H{Evalute test cases and display results}

|  u = pop[0][0];
\S ^^^^^^^^^^^^^ Temporary hack
|
|  OrgTest(u);
|
|  printf("Test results:\n");
|  for (t = 0; t < c.ntest; t++)
|  {
|    int i;
|
|    printf("   Input:");
|    for (i = 0; i < c.ni; i++)
|      printf("  %+6.3f", (double) c.test[t]->m[i].r);
|
|    printf("   Output:");
|    for (i = 0; i < c.no; i++)
|      printf("  %+8.3f", (double) c.test[t]->m[c.ni+c.nh+i].r);
|
|    printf("\n");
|  }
|  printf("\n");

\H{SPECIAL:  Test iteration of logistic equation}

|  if ((c.ni == 1) && (c.no == 1))
|  {
|    real x = 0.5;
-    real x = 0.152074;
|    long i = 0;
|
|    printf("SPECIAL ITERATION TEST:\n\n");
|
|    for (i = 0; i < 1000+100; i++)
|    {
-      if (kbhit())
|      if (0)
|      {
|        int key = getch();
|        if (key == KEY_NULL)
|          getch();
|        else if (key == KEY_ESC)
|          break;
|      }
|
|      if (i >= 1000)
|        printf("%6ld. %11.6f\n", i, x);
|
|      x = OrgIter(u, x);
|    }
|  }

\H{Destroy test organisms and exit}

|  for (i = 0; i < c.ndemes; i++)
|  for (j = 0; j < c.npop;   j++)
|    OrgDestroy(pop[i][j]);
|  DestroyArray2(pop);
|  return 0;
|}



TODD S. LEHMAN, JANUARY 1993.




