/*----------------------------------------------------------------------------*
\Module{Evolutioary Landscape}

This program tests the evolutionary landscape in N dimensions using
Monte Carlo methods.  It generates an organism with random values in
place of the "x" symbols in the config file.  Then, it tests this
organism for fitness, and starts over again.  Using this method, it is
possible to establish many points on the evolutionary landscape.

This outputs to the file "landscape.dat" in the following format:
dist	wdist	edist	M1	M2	...	B1	B2	...
*/

#include <stdlib.h>
#include <stdio.h>
//#include <conio.h>
#include <string.h>
#include <ctype.h>

#include "neural.h"
#include "protos.h"

#define N_ORGS 30000
#define M_MIN -1
#define M_MAX 1
#define B_MIN -1
#define B_MAX 1

int main (int argc, char *argv[])
{
  int i, j, k;
  Org* o;
  // Load the config file
  char *configfile;
  configfile = (char *) 0;
  
  for (i = 1; i <  argc; i++)
    if (argv[i][0] != '-')
      configfile = argv[1];

  if (!configfile)
  {
    printf("usage: landscape [options] inputfile\n");
    return 1;
  }

  if (!ReadConfiguration(configfile) |
      !ReadArgConfiguration(argc, argv))
    return 1;

  DisplayConfiguration();

  // Initialize the random seed
  RandStart(c.seed);

  // Open a file to write data to
  FILE *f;
  if ((f = fopen("landscape.dat", "w")) == NULL) 
    printf("Error, could not open output file.\n"), exit(3);
    

  // The main loop.  On each iteration, this creates an organism,
  // executes it, and checks for its fitness.  Then, it prints this
  // information to the file "landscape.dat".
  for (i=0; i<N_ORGS; i++) {
    o = OrgCreate(1);
    OrgInit(o);

    // Choose random numbers
    for (j=0; j<o->ploid[0].M->n; j++) 
      if (o->ploid[0].M->m[j].mut)
        o->ploid[0].M->m[j].r = (real)(Rand()*(M_MAX-M_MIN) + M_MIN);
    for (j=0; j<o->ploid[0].B->n; j++)
      if (o->ploid[0].B->m[j].mut)
        o->ploid[0].B->m[j].r = (real)(Rand()*(B_MAX-B_MIN) + B_MIN);
      
    // OrgEval takes us through all of the iterations necessary for
    // the organism, and puts the results in o->wdist, o->edist, and
    // o->dist.
    OrgEval(o);

    // Print out dist, e-dist, and w-dist
    fprintf(f, "%10.6f\t%10.6f\t%10.6f", (double)o->dist, (double)o->edist, (double)o->wdist);
    
    // Now print the mutable M and B values
    for (j=0; j<o->ploid[0].M->n; j++)
      if (o->ploid[0].M->m[j].mut)
        fprintf(f, "\t%10.6f", (double)o->ploid[0].M->m[j].r);
    for (j=0; j<o->ploid[0].B->n; j++) 
      if (o->ploid[0].B->m[j].mut)
        fprintf(f, "\t%10.6f", (double)o->ploid[0].B->m[j].r);
      
    fprintf(f, "\n");
    OrgDestroy(o);
  }
}
