                      Neural Network source code
                    By: Todd Lehman  and Max Shinn
                        Opened: January, 1993

* Introduction

This is a program for research on artificial neural
nets---specifically for biological applications---though the program
is general and has other uses.  The neural net has a /dynamic
structure/ comprising the instantaneous state of each neuron.  This
dynamic structure is simply a set of real numbers, one number
representing the /activity level/ of each neuron in the net.  The
dynamic structure changes with time.  The neural net also has a
/static structure/ comprising three things for each neuron:

- A threshold, also a real number,
- A set of real coefficients representing weighted stimuli arriving
  from other neurons in the net, and
- A set of real coefficients representing weighted responses sent to
  other neurons in the net.

The static structure does not change with time---or at it least
changes on a much slower time scale.  Changing the static structure is
called /training/.

Some of the neurons are designated /input neurons/; their activity
levels are simply the values of the external stimuli attached to them.
Others are designated /output neurons/; their activity levels are
taken as responses of the entire neural net and connected to an
external object or to another neural net.  The remainder are called
hidden neurons.

Among the $n$ neurons in a network are at most $n^2$ synapses and $n$
thresholds.  Synapses are represented by a matrix $M$ of weight
values, thresholds by a vector $b$, and the state of the system at
time $t$ by a vector $x_t$:

@@@
\begin{equation*}
M   = 
\begin{pmatrix}
w_{11} & w_{12} & \ldots & w_{1n} \\
w_{21} & w_{22} & \ldots & w_{2n} \\
\vdots & \vdots & \ddots & \vdots \\
w_{n1} & w_{n2} & \ldots & w_{nn} 
\end{pmatrix}
\,\,\,\,\,\,
b   = 
\begin{pmatrix} 
b_1    \\
b_2    \\
\vdots \\
b_n    \\ 
\end{pmatrix}
\,\,\,\,\,\,
x_t = \begin{pmatrix}
 x_{t,1} \\
 x_{t,2} \\
 \vdots  \\
 x_{t,n} \\ 
\end{pmatrix}
\,.
\end{equation*}
@@@

In these terms, the neural net becomes nothing more than the discrete dynamical
system defined by

$$$
x_{t+1} = F(Mx_t + b)
$$$

where

@@@
\begin{equation*}
F(x) = 
\begin{pmatrix} 
f(x_1)
f(x_2)
\vdots
f(x_n)
\end{pmatrix}
\end{equation*}
@@@

and $f(u)=\tanh(u/T)$ for hidden neurons and $f(u)=u$ for input and
output neurons.  $T$ is a simulated /temperature/---the degree of
randomness in the net.  The lower $T$, the sharper the non-linearity
in $F$.  With $T=0$, the neural net becomes a digital device.

From this definition, $w_ij$ is the strength of the synapse by which
neuron $j$ stimulates neuron $i$.  The neural net is an iterated
nonlinear map on an $n$-dimensional real space $F:ℝ^n → ℝ^n$.  All
synchronous neural nets that work from a sum of products and a
threshold are special cases of this model.

For example, consider a standard feed-forward neural net with four input
neurons, two hidden layers of three neurons each, and two output neurons.  This
is twelve neurons with only a small fraction of the $12×12$ synapses
connected.  With neurons 1--4 as the inputs and neurons 11--12 as output, the
matrix has the structure.

>                       1  2  3  4  5  6  7  8  9 10 11 12
>                    1  1  -  -  -  -  -  -  -  -  -  -  -
>                    2  -  1  -  -  -  -  -  -  -  -  -  -
>                    3  -  -  1  -  -  -  -  -  -  -  -  -
>                    4  -  -  -  1  -  -  -  -  -  -  -  -
>                    5  X  X  X  X  -  -  -  -  -  -  -  -
>                    6  X  X  X  X  -  -  -  -  -  -  -  -
>                    7  X  X  X  X  -  -  -  -  -  -  -  -
>                    8  -  -  -  -  X  X  X  -  -  -  -  -
>                    9  -  -  -  -  X  X  X  -  -  -  -  -
>                   10  -  -  -  -  X  X  X  -  -  -  -  -
>                   11  -  -  -  -  -  -  -  X  X  X  -  -
>                   12  -  -  -  -  -  -  -  X  X  X  -  -

Here, X's represents real synapses (some value $w_ij≠0$ takes the
place of each "X"), dashes ("-") represent nonexistant synapses (where
$w_ij=0$), and 1's are used to hold values on input neurons.

** Training

[Include a section on genetic search explaining that the neural nets
are trained by simulated natural selection.  The program maintains a
population of diploid matrices.  Each cell of the operational matrix
is the arithmetic mean (or perhaps the sum) of the two haploid parts
of the diploid matrix.  New matrices are generated by recombination
between pairs of diploid matrices and by mutation---using assortment
among rows of the matrix and crossover within the rows.  The genetic
algorithm is independent of the rest of the program.]

** Format of the program

The program is written in ANSI C in a document format.  Processed by
a "STAGE2" filter (Waite 1969) and further processed by the
typesetting system \TeX\ (Knuth 1986), it forms a document that
describes the program in all detail.  Processed by the utility 'd2d'
it forms a program in normal C.  In the makefile that accompanies
this program, the command 'make program' creates an executable
version of the program and 'make document' creates a printable
document.  To prepare the document, one needs 'tex' and 'stage2'; to
make the program, one needs an ansi C compiler.  The document
version is included in the files labelled '*.dvi', and a version in
a 300-dpi format is included in files labelled '*.hp'.  To print a
formatted copy of the document on a PC, simply type the command
'copy /b *.hp lpt1'.

[We're not actually using "STAGE2"---we're using 'd2c'.]

[This last section does not belong here or in the makefile, but in
an abstract or something before the top-level table of contents.]

UPDATE (11/19/12): MWS converted this program to his experimental
program for lab notes and literate programming.  The code can be
extracted with the "extract-code.pl" command, and the pdf can be
generated with the "latexify-notes" command.  Both of these should be
available as a part of the notes distribution, but if not, they can be
downloaded from [[http://code.launchpad.net/~trombonechamp/+junk/notes]].

* Sample configuration file

** Syntax

The standard form for parameters in the config file is:

> |Option Value

Any line that doesn't start with a "|" will be ignored by the
interpreter.  Matrices can also be input into the config file, though
they have slightly different structure, as shown below:

> |Matrix
> | 100000  0
> | 010000  0
> | xxxxx0  x
> | xxxxx0  x
> | xxxxx0  x
> | 00xxx0  x

The length of the matrix is predetermined by other variables in the
config file.  Anything after the expected length will be ignored.
There are no errors for matrices that are too short, so if you get
strange results, check all the parameters to make sure the matrix is
the correct length.

It is also possible, with some options, to enter multiple values for
the same option.  For instance,

> |Train 0 0  0
> |Train 0 1  1
> |Train 1 0  1
> |Train 1 1  0

** Typical values

> |RandSeed 0

The random seed is self explanatory.  It should be an integer.

> |Neurons 3

"Neurons" sets the size of the matrix, and is the total number of
neurons in the network.  See below for more information about the
matrix.

> |Input 1

"Input" is the number of input neurons.  They are assumed to be the
first neurons in the matrix.

> |Output 1

"Output" is the number of output neurons.  They are assumed to be the
last neurons in the matrix.  Consequently, the number of hidden
neurons can be determined by the equation:

$$$
\mathrm{Neurons} - (\mathrm{Input} + \mathrm{Output})
$$$

> |Ploidy 1

Ploidy is not yet fully implemented, and should be set to 1.

> |Demes 1

Multiple demes are not yet fully implemented, and should be set to 1.

> |Organisms 5
> |Population 5

Both of these options do the same thing due to a bug.  They both
define how many organisms are in each deme.  For consistency with
previously run experiments, they have both been left in the program.
However, "Organisms" is now preferred, and "Populations" is
depreciated.

> |Mutations 0.1

Every mutable matrix element has this probability of being mutated.
In theory, sometimes networks may never be mutated, and sometimes
every element in the matrix may be mutated, since it is not on a
``mutations-per-organism'' quantity.

> |PZero 0

This is the probability of any given matrix element being set to zero,
given that the matrix element is already being mutated.  Increasing
this to a non-zero number (such as $0.1$ or $0.05$) helps networks to
evolve faster, and creates cleaner networks.  The probability that any
given matrix element will be set to zero is thus given by:

$$$
\mathrm{PZero} × \mathrm{Mutations}
$$$

> |Iterations 2

This is the number of times a network will be processed before the
results are recorded.  In feed-forward networks, it makes no
difference what this number is as long as it is strictly greater than
the number of hidden layers.  In fully-connected networks, larger
values are necessary to create bistable memory systems, but
simulations of simple function regression has shown smaller values
tend to be better for simple tasks.

> |InitZero 0

This boolean determines whether the network be initialized to random
values, or to all be zero.  Random gives faster evolution, though zero
might be more biologically relevant.

> |DiffEq 0

Not yet implemented... should be set to 0.

> |NormalizeAll 1

"NormalizeAll" dictates whether all neruons should be normalized with
the arctan normalization function, or whether only hidden neurons
should be normalized.  If this is set to 1, the output of the network
is constrained to values between 0 and 1.  

> |StepNormalize 1

This boolean sets the normalization function to be a Heaviside step
function instead of arctan.  The primary purpose of this is to develop
boolean gates.  As I understand it (though this may not be correct),
enabling this option turns the neural network into a Markov network.

> |SynapseAlpha .1
> |SynapseBeta 2

These options determine the added synaptic error, "w-dist".
"SynapseAlpha" is the scalar, and "SynapseBeta" is the exponent.

$$$
α ∑_{i=1}^n ∑_{j=1}^n |a_ij|^β
$$$

> |CostOnB 1

This boolean determines whether or not the threshold values are
subjected to the above calculation.  If 0, they are excluded from this
calculation.  Biologically speaking, enabling this option is analogous
to having a selective disadvantage for neuron threshold values.

> |NormWeights 1

This boolean flag determines whether or not the weights should be
normalized.  In other words, should the w-dist be divided by the
number of synapses?  The default is true for historical compatibility,
and it is probably a good idea to keep it here unless you're trying to
compare convergence amonst networks of different structures.

> |NoiseSigma 0

If this is set to zero (the default), there will be no noise in the
training data.  Otherwise, a normally distributed random variable with
this standard deviation and a mean of 0 will be added to each training
data point.  This random variable will be regenerated for each
organism in each generation.

> |Matrix
> | 100000  0
> | 010000  0
> | xxxxx0  x
> | xxxxx0  x
> | xxxxx0  x
> | 00xxx0  x

"Matrix" is the connection matrix.  An element $a_ij$ will synapse
from neuron $j$ onto neuron $i$.  The last column specifies the
threshold of neuron $i$.  Thus, the matrix should be $(N+1)×N$, where
$N=\mathrm{Neurons}$.

A "1" signifies a synapse will always have the value 1, i.e. the full
firing rate will be directly added onto the postsynaptic neuron.  This
should generally be used for input neurons onto themselves, because if
they have no other incoming connections, it ensures they will continue
to hold the same value over time.  A value of "0" means that no
synapse will ever occur between these two neurons.  A value of "x"
specifies that this value is mutable, and will be adjusted over time
as the population evolves.  

> |Train 0 1  0
> |Train 0 1  1
> |Train 1 0  1
> |Train 1 1  0

There may be any number of "Train" statements.  They should have a
number of arguments equal to the sum of the input and output neurons.
The first set of arguments is any given input to the network, and the
final arguments are the network's expected output.  In the example
given here, there are two inputs and an output, training the network
to perform an XOR.

> |Test .2 .2
> |Test .4 .4
> |Test .6 .6
> |Test .8 .8

"Test" is very similar to train, except the results are used to test
the network after the training period.  The number of arguments should
be equal to the number of input neurons.  These values are fed into
the network, and the output is displayed at the end of the training
period.  It is also written to the data file if one was specified.

* Implementation

** Header files

First we include the header files.

===>
#include <stdlib.h>
#include <stdio.h>
//#include <conio.h>
#include <string.h>
#include <ctype.h>

#include "neural.h"  // Defines all structs, #defines, and typedefs
#include "protos.h"  // Contains function prototypes for all of TSL's functions
<===

** Main

Command-line switches are defined in the configuration module.  A file
name is expected as the first non-switch argument, which will serve as
the configuration file.

*Entry conditions*

- "argc" specifies the number of arguments to the program.
- "argv" is a list of pointers to the arguments.

*Exit conditions*

"main" returns a result code:

- 0  No problems were encountered.
- 1  Problems were encountered with the command line.
- 2  A memory allocation failure occurred.


===>
// #define  MAX_ORGANISMS  100 // (I don't think this is needed. -MWS 2012)

int main (int argc, char *argv[])
{
  printf("________________________________________"
         "_______________________________________\n\n");
<===

** Main's variables

The organism "u" is used for a bunch of miscellaneous purposes as a
generic pointer to an organism.  The organism "v" is the organism
currently being mutated and evolved.  "pop" is the population of
organisms collectively being evolved.  

-> Note that even though "pop" implies ``population'', there are many
populations within each deme, and many demes in the "pop".  Thus,
"pop" misleadingly contains many populations.
"pop[ndeme][npopulation]" is the normal method for accessing an
organism, where "ndeme" is the organism's deme (i.e. always 0 if there
is only one deme), and "npopulation" is the organism's index within
the deme.

===>
  Org *u, *v, **(*pop);
<===

"configfile" is the string that gives the file name of the config file
passed as a command line parameter.

===>
  char *configfile;
<===

"gen" is the current generation.  Once "gen" is equal to the config
file option "Generations", the simulation is over.

===>
  long gen;
<===

"i", "j", and "t" are used for counting.  "t" is only used in the test
loop, while "i" and "j" are used generically.

===>
  int i, j, t;
<===

** Load the configuration file

First, scan for the name of the config file and store it in
"configfile".  Exit if none was given at the command line.

===>
  configfile = (char *) 0;

  for (i = 1; i < argc; i++)
    if (argv[i][0] != '-')
      configfile = argv[1];

  if (!configfile)
  {
    printf("usage: neural [options] inputfile\n");
    return 1;
  }
<===

Now read it into memory.  Its parameters can be accessed with the
variable "c".

===>
  if (!ReadConfiguration(configfile) |
      !ReadArgConfiguration(argc, argv))
    return 1;

  DisplayConfiguration();
<===

Set the random seed immediately so the results are perfectly
reproducible.

===>
  RandStart(c.seed);
<===

** Begin the evolution loop

Go through and fill up "pop" with organisms.  The first index on "pop"
is which deme it is in, and the second number is the index within the
deme.

===>
  pop = (Org ***) CreateArray2(c.ndemes, c.npop, sizeof(Org *));
  for (i = 0; i < c.ndemes; i++)
  for (j = 0; j < c.npop;   j++)
  {
    Org *o = OrgCreate(c.ploidy);
    OrgInit(o);

    if (c.initcoords != NULL)
      OrgFromCoords(o, c.initcoords);
    else if (c.initzero)
      OrgZero(o);
    else
      OrgRand(o);

    OrgEval(o);

    pop[i][j] = o;                         // ** I think I want OrgInsert here.
  }
<===

** Open the trace file

Open the file to save the trace.  The filename contains the
random seed so that multiple instances can be run at the same
time on the same computer and not overwrite each other.  The
trace is described in pathtracing.n and allows one to track
changes in the most fit member of the population

===>
  FILE *f;
  char *fn = malloc(255*sizeof(char));
  sprintf(fn, "trace-%i.csv", (int)c.seed);
  if ((f = fopen(fn, "w")) == NULL) 
    printf("Error, could not open trace output file.\n"), exit(3);
<===

This opens the file to save the output function to a at every frame
where it evolves.  Unless you want to make an animated gif of the
output function, you shouldn't need this, and may comment it out.
Also comment out the section below (C-s ff,).

===>
  // if ((ff = fopen("trace-function.csv", "w")) == NULL)
  //    printf("error, could not open trace function output file.\n"), exit(3);
<===

** Evolve the population until the user signals termination of the program

This segment is responsible for running through each generation, where
the population improves or maintains its fitness every generation.  

The following lines were written so that the user could have some
real-time control during the program's execution.  They only worked on
DOS, and don't seem to work now (GCC 4.6, Linux 3.2).  I'm leaving it
in for historical reasons.

===>
  #define  KEY_NULL   0
  #define  KEY_ESC    27
  #define  KEY_SPACE  32

  for (gen = 0; (!c.ngen || (gen < c.ngen)); gen++)
  {
//    if (kbhit())                           // Check the keyboard and read a
    if (0)                                 // Check the keyboard and read a
    {                                      // character if one is availble
      int key = getch();                   // (MS-DOS) specific).

      if (key == KEY_NULL)                 // If the special null code is read,
      {                                    // throw away the next character
        getch();                           // without translation.  (MS-DOS)
      }
      else if (key == KEY_ESC)             // If the Escape key is pressed, halt.
      {
        break;
      }
      else if (key == KEY_SPACE)           // If the space bar is pressed,
      {                                    // display the current weight and
        MatDisplay(pop[0][0]->ploid[0].M, "M"); //  threshold matrices.
        VecDisplay(pop[0][0]->ploid[0].B, "B", VDISPLAY);
//                                   ^ Temporary hack
      }
    }
<===

Each deme is evolved seperately.  A random organism from each deme is
selected.  A copy of that organism is subjected to random mutations
via the OrgMutate function.  Then, if this new mutilated organism
fares better than the worst organism in the deme, it is added to the
deme (via OrgInsert). 

-> Currently, there is no advantage to using more than one deme, since
only the first deme is reported to the user or logged in the data
files.

===>
    for (i = 0; i < c.ndemes; i++)
    {
      u = pop[i][RAND_INT_RANGE(0, c.npop - 1)];
      v = OrgClone(u);
      if (c.sigma != 0)
        PopEvalNoise(pop[i]);
      OrgMutate(v);
      OrgEval(v);
<===

If the organism is inserted into the first deme (deme 0), print out
information about it, and write this event to the trace file.

===>
      if (OrgInsert(pop[i], c.npop, v) == 0)
      {
        if (c.errorweight)
          printf("Gen:%9ld      "
                 "e-dist:%10.6f     w-dist:%10.6f     dist:%10.6f\n",
                 gen, (double)v->edist, (double)v->wdist, (double)v->dist);
        else
          printf("Generation: %9ld            "
                 "e-dist: %13.6f\n",
                 gen, (double)v->edist);
<===

Also, write information about this organism to the trace file.  This
is in the format:

> generation, dist, edist, wdist, m_1, m_2, ..., b_1, b_2, ...

E-dist is the mean squared error of the current network as it compares
to the training data (i.e. ``error distance'').  The w-dist is the
mean value of the synapse connections.

-> This may not be the best measurement for every situation,
especially of networks of different sizes.  It might be necessary to
adjust the w-dist when comparing networks with unlike structures or
numbers of synapses.  

===>
        OrgTest(v);
        fprintf(f, "%i\t%.6f\t%.6f\t%.6f", 
                 (int)gen, (double)v->dist, (double)v->edist, (double)v->wdist);
        for (j=0; j<v->ploid[0].M->n; j++)
          if (v->ploid[0].M->m[j].mut)
              fprintf(f, "\t%.6f", (double)v->ploid[0].M->m[j].r);
        for (j=0; j<v->ploid[0].B->n; j++) 
          if (v->ploid[0].B->m[j].mut)
            fprintf(f, "\t%.6f", (double)v->ploid[0].B->m[j].r);
        fprintf(f, "\n");
<===

This will print out the organism's best fit function, with a data
point at each "Test" point.  It will add to this data file each time
an organism is inserted into the first deme.  This can be used to show
the best fit line at any given time.  If not generating a gif, this
should probably be commented out.

===>
        // int t;
        // for (t=0; t<c.ntest; t++)
        //   for (j=0; j<c.no; j++)
        //     fprintf(ff, "%f\t%i\t%.6f\n", c.test[t]->m[j].r, j, (double) c.test[t]->m[c.ni+c.nh+j].r);
        // fprintf(ff, "\n");
//      //   OrgHolisticTest(v);
      }
<===

No matter what, after every 5000 generations, print the structure of
the new mutant.

===>
      if (gen % 5000 == 0)
      {
        if (c.errorweight)
          printf("Gen:%9ld\n", gen);
        else                                    // <<-- yuck, clean this up!
          printf("Generation: %9ld\n", gen);

        MatDisplay(v->ploid[0].M, "M"); //  threshold matrices.
        VecDisplay(v->ploid[0].B, "B", VDISPLAY);
      }
    }
  }
<===

** Report final matrix.

Since organisms are arranged by fitness within demes, printing the
first organism in the deme will print the one with the highest
fitness.  Again, using multiple demes isn't fully implemented, so
we're only displaying the the most fit organism from the first deme.

===>
  MatDisplay(pop[0][0]->ploid[0].M, "M");
  VecDisplay(pop[0][0]->ploid[0].B, "B", VDISPLAY);
//                              ^ Temporary hack
<===

** Evalute test cases and display results

Again, using the most fit organism from the first deme, we find the
organism's response to the test data.  Then, we display the results.

===>
  u = pop[0][0];
//  ^^^^^^^^^^^^^ Temporary hack

  OrgTest(u);

  printf("Test results:\n");
  for (t = 0; t < c.ntest; t++)
  {
    int i;

    printf("   Input:");
    for (i = 0; i < c.ni; i++)
      printf("  %+6.3f", (double) c.test[t]->m[i].r);

    printf("   Output:");
    for (i = 0; i < c.no; i++)
      printf("  %+8.3f", (double) c.test[t]->m[c.ni+c.nh+i].r);

    printf("\n");
  }
  printf("\n");
<===

** Quantify the hypernormalness

This will measure the degree to which the network is hypernormal,
i.e. it doesn't follow a strictly increasing/decreasing extrapolation,
whether sigmoidally, linearly, or in any other form.  I call this
value Ξ, for the shallow reason that its measurement method is
remeniscent of its shape.  This currently only works with networks
with one input and one output.

Ξ is calculated as follows:

- Find the curve of the function generated by the network's test
  cases.
- For each test case's coordinate (input, output), draw a horizontal
  line across the graph and count the number of times it intersects
  the line on the graph.
- Take the mean of all of the test cases.

In networks that are strictly increasing or decreasing, this number
will be one.  However, it should grow quickly in networks that are
very hypernormal.  Monotonic functions may have a Ξ less than 1.  A
horizontal line will have $Ξ = 0$.

Let us say $t_i$ is the $i$'th test case input, and $f(t_i)$ is the
$i$'th test case's output.  Thus, by taking the sum of boolean
variables, we find define Ξ as

$$$
Ξ = 1/(N-1) ∑_{j=1}^i ∑_{k=1}^{i-1} ((f(t_{k+1})≥f(t_j)>f(t_k)) ∨ (f(t_{k+1})<f(t_j)≤f(t_k)))
$$$

The algorithm here is $Ω(n^2)$.  It could probably be made more
efficient, though since it is only executed once, this shouldn't be a
problem.

This is guarded to ensure Ξ is relevant and will not cause a division
by zero error.  In the case that this routine is not executed, Ξ will
be set to $-1$.

===>
real xi = -1;
if (c.ntest > 1 && c.ni == 1 && c.no == 1) {
  int hyperness = 0;
  for (i=0; i<c.ntest; i++) {
    for (j=0; j<c.ntest-1; j++) {
      real testbase = c.test[i]->m[c.ni+c.nh].r;
      real testabove = c.test[j+1]->m[c.ni+c.nh].r;
      real testbelow = c.test[j]->m[c.ni+c.nh].r;
      if ((testabove >= testbase && testbase > testbelow) ||
          (testbelow >= testbase && testbase > testabove))
        hyperness += 1;
    }
  }
  // Each iteration through the loop of j's is guaranteed to hit at
  // least one neuron, so we divide by (c.ntest-1).
  xi = (real)hyperness/(c.ntest-1);
  printf("\n\nXi score is: %f\n\n", xi);
}
<===

** Write the test results to the datafile

Here, we open the output file, if one was specified in the config file
or on the command line.  If the file doesn't exist, we create a
header.  Then, we write out all of the settings used in the network,
followed by the results of the test data calculated previously.

===>
  OrgEval(u);
  if (c.outfile != NULL) {
    // If the file exists
    FILE *file = fopen(c.outfile, "r");
    int i;
    // File doesn't exist.  Open the file for writing and write the header
    if (file == NULL) {
      file = fopen(c.outfile, "a+");
      if (file == NULL) 
        printf("Could not open output file for writing.\n"), exit(1);
      // Write the csv header if the file doesn't exist
      fprintf(file, "RandSeed,Neurons,Ni,No,Ntrain,Ntest,Ploidy,Demes,PZero,"
                    "Population,Generations,Mutations,Iterations,Initzero,"
                    "Diffeq,Normall,Errorweight,Alpha,Beta,NoiseSigma,dist,"
                    "edist,wdist,Xi");
      for (i=0; i<c.ni; i++)
        for (j=0; j<c.ntest; j++)
          fprintf(file, ",Input_%i_Test_%i", i+1, j+1);
      for (i=0; i<c.no; i++)
        for (j=0; j<c.ntest; j++)
          fprintf(file, ",Output_%i_Test_%i", i+1, j+1);
      fprintf(file, "\n");
    }
    // File exists.  Open the file for writing.
    else {
      fclose(file);
      file = fopen(c.outfile, "a+");
      if (file == NULL)
        printf("Could not open output file for writing.\n"), exit(1);
    }
    
    //Write the test case data
    fprintf(file, "%i,%i,%i,%i,%i,%i,%i,%i,%.6f,%i,%i,%.6f,%i,%i,%i,%i,%i,%.6f,"
                  "%.6f,%.6f,%.6f,%.6f,%.6f,%.6f", 
            (int)c.seed,c.nn,c.ni,c.no,c.ntrain,c.ntest,c.ploidy,c.ndemes,c.pzero,
            (int)c.npop,(int)c.ngen,c.pmut,c.niter,c.initzero,c.diffeq,c.normall,
            c.errorweight,c.walpha,c.wbeta,c.sigma,u->dist,u->edist,u->wdist,xi);
    for (j=0; j<c.ni; j++)
      for (i=0; i<c.ntest; i++)
        fprintf(file, ",%.6f", (double) c.test[i]->m[j].r);
    for (j=0; j<c.no; j++)
      for (i=0; i<c.ntest; i++)
        fprintf(file, ",%.6f", (double) c.test[i]->m[c.ni+c.nh+j].r);
    fprintf(file, "\n");

    fclose(file);
  }
<===

** Special Test iteration of logistic equation

I still don't know what this is for.  Leaving it in anyway in case it
does something we may need in the future.

===>
  if ((c.ni == 1) && (c.no == 1))
  {
    real x = 0.5;
//    real x = 0.152074;
    long i = 0;

    printf("SPECIAL ITERATION TEST:\n\n");

    for (i = 0; i < 1000+100; i++)
    {
//      if (kbhit())
      if (0)
      {
        int key = getch();
        if (key == KEY_NULL)
          getch();
        else if (key == KEY_ESC)
          break;
      }

      if (i >= 1000)
        printf("%6ld. %11.6f\n", i, x);

      x = OrgIter(u, x);
    }
  }
<===

** Destroy test organisms and exit

Free all the memory allocated in "pop" and exit successfully.

===>
  for (i = 0; i < c.ndemes; i++)
  for (j = 0; j < c.npop;   j++)
    OrgDestroy(pop[i][j]);
  DestroyArray2(pop);
  return 0;
}
<===

